from __future__ import annotations

from .models import ControlState, NightLightState, SensorState, SettingsState, StatusState
from .proto import (
    ControlNightLight,
    MountingMode,
    Response,
    SensorData,
    SettingsWifiBand,
    StatusConnectionToServer,
)

_WIFI_BAND_MAP: dict[int, str] = {
    SettingsWifiBand.ANY: "any",
    SettingsWifiBand.FR2_4GHZ: "2.4ghz",
    SettingsWifiBand.FR5_0GHZ: "5ghz",
}

_MOUNTING_MODE_MAP: dict[int, str] = {
    MountingMode.STAND: "stand",
    MountingMode.TRAVEL: "travel",
    MountingMode.SWITCH: "switch",
}


def _parse_sensor_data(
    sensor_data_list: list[SensorData],
    current: SensorState,
) -> SensorState:
    from .proto import SensorType as ProtoSensorType

    temperature = current.temperature
    humidity = current.humidity
    light = current.light
    sound_alert = current.sound_alert
    motion_alert = current.motion_alert
    night = current.night

    for sd in sensor_data_list:
        if sd.sensor_type == ProtoSensorType.TEMPERATURE:
            if sd.value_milli:
                temperature = sd.value_milli / 1000.0
            elif sd.value:
                temperature = float(sd.value)
        elif sd.sensor_type == ProtoSensorType.HUMIDITY:
            if sd.value_milli:
                humidity = sd.value_milli / 1000.0
            elif sd.value:
                humidity = float(sd.value)
        elif sd.sensor_type == ProtoSensorType.LIGHT:
            light = sd.value
        elif sd.sensor_type == ProtoSensorType.SOUND:
            sound_alert = sd.is_alert
        elif sd.sensor_type == ProtoSensorType.MOTION:
            motion_alert = sd.is_alert
        elif sd.sensor_type == ProtoSensorType.NIGHT:
            night = bool(sd.value)

    return SensorState(
        temperature=temperature,
        humidity=humidity,
        light=light,
        sound_alert=sound_alert,
        motion_alert=motion_alert,
        night=night,
    )


def _parse_status(resp: Response) -> StatusState:
    if resp.HasField("status"):
        return _parse_status_from_proto(resp.status)
    return StatusState()


def _parse_status_from_proto(status: object) -> StatusState:
    from .proto import Status as ProtoStatus

    if not isinstance(status, ProtoStatus):
        return StatusState()

    connected: bool | None = None
    if status.HasField("connection_to_server"):
        connected = status.connection_to_server == StatusConnectionToServer.CONNECTED

    return StatusState(
        connected_to_server=connected,
        firmware_version=status.current_version or None,
        hardware_version=status.hardware_version or None,
        mounting_mode=_MOUNTING_MODE_MAP.get(status.mode),
    )


def _parse_settings(resp: Response) -> SettingsState:
    if resp.HasField("settings"):
        return _parse_settings_from_proto(resp.settings)
    return SettingsState()


def _parse_settings_from_proto(settings: object) -> SettingsState:
    from .proto import Settings as ProtoSettings

    if not isinstance(settings, ProtoSettings):
        return SettingsState()

    return SettingsState(
        night_vision=settings.night_vision if settings.HasField("night_vision") else None,
        volume=settings.volume if settings.HasField("volume") else None,
        sleep_mode=settings.sleep_mode if settings.HasField("sleep_mode") else None,
        status_light_on=settings.status_light_on if settings.HasField("status_light_on") else None,
        mic_mute_on=settings.mic_mute_on if settings.HasField("mic_mute_on") else None,
        wifi_band=_WIFI_BAND_MAP.get(settings.wifi_band)
        if settings.HasField("wifi_band")
        else None,
        mounting_mode=_MOUNTING_MODE_MAP.get(settings.mounting_mode)
        if settings.HasField("mounting_mode")
        else None,
    )


def _parse_control(resp: Response) -> ControlState:
    if resp.HasField("control"):
        return _parse_control_from_proto(resp.control)
    return ControlState()


def _parse_control_from_proto(control: object) -> ControlState:
    from .proto import Control as ProtoControl

    if not isinstance(control, ProtoControl):
        return ControlState()

    night_light: NightLightState | None = None
    if control.HasField("night_light"):
        if control.night_light == ControlNightLight.LIGHT_ON:
            night_light = NightLightState.ON
        else:
            night_light = NightLightState.OFF

    sensor_transfer_enabled: bool | None = None
    if control.HasField("sensor_data_transfer"):
        sdt = control.sensor_data_transfer
        sensor_transfer_enabled = any(
            [sdt.sound, sdt.motion, sdt.temperature, sdt.humidity, sdt.light, sdt.night]
        )

    return ControlState(
        night_light=night_light,
        night_light_timeout=control.night_light_timeout
        if control.HasField("night_light_timeout")
        else None,
        sensor_data_transfer_enabled=sensor_transfer_enabled,
    )



from __future__ import annotations

import asyncio
import time
from unittest.mock import AsyncMock, MagicMock, patch

import aiohttp
import pytest

from aionanit.auth import TokenManager
from aionanit.camera import GetStatus, NanitCamera, RequestType, Response
from aionanit.exceptions import NanitTransportError
from aionanit.models import TransportKind
from aionanit.rest import NanitRestClient


def _make_camera() -> tuple[NanitCamera, MagicMock]:
    session = MagicMock(spec=aiohttp.ClientSession)
    rest = MagicMock(spec=NanitRestClient)
    token_manager = MagicMock(spec=TokenManager)
    token_manager.async_get_access_token = AsyncMock(return_value="test_token")

    camera = NanitCamera(
        uid="cam_uid_1",
        baby_uid="baby_uid_1",
        token_manager=token_manager,
        rest_client=rest,
        session=session,
        prefer_local=False,
    )
    return camera, token_manager


@pytest.mark.asyncio
async def test_token_refresh_task_started_on_start() -> None:
    camera, token_manager = _make_camera()
    token_manager._expires_at = time.monotonic() + 3600.0

    camera._transport = MagicMock()
    camera._transport.async_connect_cloud = AsyncMock()
    camera._transport.transport_kind = TransportKind.CLOUD
    camera._transport.connected = True

    camera._async_request_initial_state = AsyncMock()
    camera._async_enable_sensor_push = AsyncMock()
    camera._start_health_check = MagicMock()
    camera._start_sensor_poll = MagicMock()

    await camera.async_start()

    assert camera._token_refresh_task is not None
    assert not camera._token_refresh_task.done()

    camera._cancel_token_refresh()


@pytest.mark.asyncio
async def test_token_refresh_task_cancelled_on_stop() -> None:
    camera, token_manager = _make_camera()
    token_manager._expires_at = time.monotonic() + 3600.0

    camera._transport = MagicMock()
    camera._transport.async_close = AsyncMock()

    camera._start_token_refresh()
    refresh_task = camera._token_refresh_task

    await camera.async_stop()
    await asyncio.sleep(0)

    assert camera._token_refresh_task is None
    assert refresh_task is not None
    assert refresh_task.cancelled()


@pytest.mark.asyncio
async def test_token_refresh_forces_reconnect_before_expiry() -> None:
    camera, token_manager = _make_camera()
    camera._stopped = False
    token_manager._expires_at = 1050.0

    camera._transport = MagicMock()
    camera._transport.connected = True

    async def _force_and_stop() -> None:
        camera._stopped = True

    camera._transport.async_force_reconnect = AsyncMock(side_effect=_force_and_stop)

    with (
        patch("aionanit.camera.time.monotonic", return_value=1000.0),
        patch("aionanit.camera.asyncio.sleep", AsyncMock(return_value=None)),
    ):
        await asyncio.wait_for(camera._token_refresh_loop(), timeout=0.1)

    camera._transport.async_force_reconnect.assert_awaited_once()


@pytest.mark.asyncio
async def test_send_request_waits_for_reconnection() -> None:
    camera, _ = _make_camera()

    camera._transport = MagicMock()
    camera._transport.connected = False
    camera._transport.idle_seconds = 0.0

    response = Response(status_code=200)

    async def _fake_send(_: bytes) -> None:
        camera._pending.resolve(1, response)

    camera._transport.async_send = AsyncMock(side_effect=_fake_send)
    camera._async_reconnect = AsyncMock()
    camera._connected_event.clear()

    async def _simulate_reconnect() -> None:
        await asyncio.sleep(0.01)
        camera._transport.connected = True
        camera._connected_event.set()

    reconnect_task = asyncio.create_task(_simulate_reconnect())

    result = await camera._send_request(
        RequestType.GET_STATUS,
        get_status=GetStatus(all=True),
    )
    await reconnect_task

    assert result.status_code == 200
    camera._async_reconnect.assert_not_awaited()


@pytest.mark.asyncio
async def test_send_request_timeout_when_not_reconnecting() -> None:
    camera, _ = _make_camera()

    camera._transport = MagicMock()
    camera._transport.connected = False
    camera._transport.idle_seconds = 0.0
    camera._transport.async_send = AsyncMock()
    camera._connected_event.clear()
    camera._connected_event.wait = AsyncMock(side_effect=TimeoutError)
    camera._async_reconnect = AsyncMock(side_effect=NanitTransportError("reconnect failed"))

    with pytest.raises(NanitTransportError, match="reconnect failed"):
        await camera._send_request(
            RequestType.GET_STATUS,
            get_status=GetStatus(all=True),
        )

    camera._async_reconnect.assert_awaited_once()
